@@ -15,7 +15,7 @@ use Mail::DKIM::Signature;
 package Mail::DKIM::Common;
 use base "Mail::DKIM::MessageParser";
 use Carp;
-our $VERSION = '0.28';
+our $VERSION = '0.30';
 
 sub new
 {
@@ -28,6 +28,11 @@ sub add_header
 	my $self = shift;
 	my ($line) = @_;
 
+	foreach my $algorithm (@{$self->{algorithms}})
+	{
+		$algorithm->add_header($line);
+	}
+
 	if ($line =~ /^([^:]+)\s*:(.*)/s)
 	{
 		my $field_name = lc $1;
@@ -135,7 +140,7 @@ sub message_originator
 		my @list = Mail::Address->parse($self->{headers_by_name}->{from});
 		return $list[0];
 	}
-	return undef;
+	return Mail::Address->new;
 }
 
 sub message_sender
@@ -153,7 +158,7 @@ sub message_sender
 		my @list = Mail::Address->parse($self->{headers_by_name}->{from});
 		return $list[0];
 	}
-	return undef;
+	return Mail::Address->new;
 }
 
 sub result
