@@ -89,7 +89,7 @@ is written to the referenced string or file handle.
 package Mail::DKIM::Verifier;
 use base "Mail::DKIM::Common";
 use Carp;
-our $VERSION = '0.29';
+our $VERSION = '0.30';
 
 sub init
 {
@@ -131,12 +131,40 @@ sub handle_header
 
 	if (lc($field_name) eq "dkim-signature")
 	{
-		$self->add_signature($line);
+		eval
+		{
+			my $signature = Mail::DKIM::Signature->parse($line);
+			$self->add_signature($signature);
+		};
+		if ($@)
+		{
+			# the only reason an error should be thrown is if the
+			# signature really is unparse-able
+
+			# otherwise, invalid signatures are caught in finish_header()
+
+			chomp (my $E = $@);
+			$self->{signature_reject_reason} = $E;
+		}
 	}
 
 	if (lc($field_name) eq "domainkey-signature")
 	{
-		$self->add_signature_dk($line);
+		eval
+		{
+			my $signature = Mail::DKIM::DkSignature->parse($line);
+			$self->add_signature($signature);
+		};
+		if ($@)
+		{
+			# the only reason an error should be thrown is if the
+			# signature really is unparse-able
+
+			# otherwise, invalid signatures are caught in finish_header()
+
+			chomp (my $E = $@);
+			$self->{signature_reject_reason} = $E;
+		}
 	}
 }
 
@@ -144,47 +172,28 @@ sub add_signature
 {
 	my $self = shift;
 	croak "wrong number of arguments" unless (@_ == 1);
-	my ($contents) = @_;
+	my ($signature) = @_;
 
-	eval
+	push @{$self->{signatures}}, $signature;
+
+	unless ($self->check_signature($signature))
 	{
-		my $signature = Mail::DKIM::Signature->parse($contents);
-		push @{$self->{signatures}}, $signature;
-	};
-	if ($@)
-	{
-		# the only reason an error should be thrown is if the
-		# signature really is unparse-able
-
-		# otherwise, invalid signatures are caught in finish_header()
-
-		chomp (my $E = $@);
-		$self->{signature_reject_reason} = $E;
+		$signature->result("invalid",
+			$self->{signature_reject_reason});
+		return;
 	}
-}
 
-# parses a DomainKeys-type signature
-sub add_signature_dk
-{
-	my $self = shift;
-	croak "wrong number of arguments" unless (@_ == 1);
-	my ($contents) = @_;
+	# create a canonicalization filter and algorithm
+	my $algorithm_class = $signature->get_algorithm_class(
+				$signature->algorithm);
+	my $algorithm = $algorithm_class->new(
+				Signature => $signature,
+				Debug_Canonicalization => $self->{Debug_Canonicalization},
+			);
 
-	eval
-	{
-		my $signature = Mail::DKIM::DkSignature->parse($contents);
-		push @{$self->{signatures}}, $signature;
-	};
-	if ($@)
-	{
-		# the only reason an error should be thrown is if the
-		# signature really is unparse-able
-
-		# otherwise, invalid signatures are caught in finish_header()
-
-		chomp (my $E = $@);
-		$self->{signature_reject_reason} = $E;
-	}
+	# save the algorithm
+	$self->{algorithms} ||= [];
+	push @{$self->{algorithms}}, $algorithm;
 }
 
 sub check_signature
@@ -198,8 +207,8 @@ sub check_signature
 		# unsupported version
 		if (defined $signature->version)
 		{
-			$self->{signature_reject_reason} = "unsupported version (v="
-				. $signature->version . ")";
+			$self->{signature_reject_reason} = "unsupported version "
+				. $signature->version;
 		}
 		else
 		{
@@ -263,28 +272,6 @@ sub check_signature
 		return 0;
 	}
 
-	unless (check_signature_identity($signature))
-	{
-		$self->{signature_reject_reason} = "bad identity";
-		return 0;
-	}
-
-	# check domain against message From: and Sender: headers
-#	my $responsible_address = $self->message_originator;
-#	if (!$responsible_address)
-#	{
-#		# oops, no From: or Sender: header
-#		die "No From: or Sender: header";
-#	}
-#
-#	my $senderdomain = $responsible_address->host;
-#	my $sigdomain = $signature->domain;
-#	if (!match_subdomain($senderdomain, $sigdomain))
-#	{
-#		$self->{signature_reject_reason} = "unmatched domain";
-#		return 0;
-#	}
-
 	return 1;
 }
 
@@ -299,13 +286,22 @@ sub check_public_key
 	{
 		$@ = undef;
 
+		# HACK- I'm indecisive here about whether I want the
+		# check_foo functions to return false or to "die"
+		# on failure
+
 		# check public key's allowed hash algorithms
 		$result = $public_key->check_hash_algorithm(
 				$signature->hash_algorithm);
 
+		# HACK- DomainKeys signatures are allowed to have an empty g=
+		# tag in the public key
+		my $empty_g_means_wildcard = $signature->isa("Mail::DKIM::DkSignature");
+
 		# check public key's granularity
 		$result &&= $public_key->check_granularity(
-				$signature->identity);
+				$signature->identity, $empty_g_means_wildcard);
+
 		die $@ if $@;
 	};
 	if ($@)
@@ -326,7 +322,7 @@ sub check_signature_identity
 
 	my $d = $signature->domain;
 	my $i = $signature->identity;
-	if ($i =~ /\@(.*)$/)
+	if (defined($i) && $i =~ /\@(.*)$/)
 	{
 		return match_subdomain($1, $d);
 	}
@@ -339,7 +335,7 @@ sub match_subdomain
 	my ($subdomain, $superdomain) = @_;
 
 	my $tmp = substr(".$subdomain", -1 - length($superdomain));
-	return (".$superdomain" eq $tmp);
+	return (lc ".$superdomain" eq lc $tmp);
 }
 
 #
@@ -360,15 +356,20 @@ sub finish_header
 		return;
 	}
 
+	foreach my $algorithm (@{$self->{algorithms}})
+	{
+		$algorithm->finish_header;
+	}
+
 	# For each parsed signature, check it for validity. If none are valid,
 	# our result is "invalid" and our result detail will be the reason
 	# why the last signature was invalid.
 
-	my @valid = ();
 	foreach my $signature (@{$self->{signatures}})
 	{
-		unless ($self->check_signature($signature))
+		unless (check_signature_identity($signature))
 		{
+			$self->{signature_reject_reason} = "bad identity";
 			$signature->result("invalid",
 				$self->{signature_reject_reason});
 			next;
@@ -385,7 +386,8 @@ sub finish_header
 			my $E = $@;
 			chomp $E;
 			$self->{signature_reject_reason} = $E;
-			$signature->result("invalid", $E);
+			$signature->result("invalid",
+				$self->{signature_reject_reason});
 			next;
 		}
 
@@ -395,42 +397,21 @@ sub finish_header
 				$self->{signature_reject_reason});
 			next;
 		}
-
-		# this signature is ok
-		push @valid, $signature;
 	}
 
-	unless (@valid)
-	{
-		# no valid signatures found
-		$self->{result} = "invalid";
-		$self->{details} = $self->{signature_reject_reason};
-		return;
-	}
-
-	# now, for each valid signature, create an "algorithm" object which
-	# will process the message
-
-	$self->{algorithms} = [];
-	foreach my $signature (@valid)
-	{
-		# create a canonicalization filter and algorithm
-		my $algorithm_class = $signature->get_algorithm_class(
-					$signature->algorithm);
-		my $algorithm = $algorithm_class->new(
-					Signature => $signature,
-					Debug_Canonicalization => $self->{Debug_Canonicalization},
-				);
-
-		# output header as received so far into canonicalization
-		foreach my $line (@{$self->{headers}})
+	# stop processing signatures that are already known to be invalid
+	@{$self->{algorithms}} = grep
 		{
-			$algorithm->add_header($line);
-		}
-		$algorithm->finish_header;
+			my $sig = $_->signature;
+			!($sig->result && $sig->result eq "invalid");
+		} @{$self->{algorithms}};
 
-		# save the algorithm
-		push @{$self->{algorithms}}, $algorithm;
+	if (@{$self->{algorithms}} == 0
+		&& @{$self->{signatures}} > 0)
+	{
+		$self->{result} = $self->{signatures}->[0]->result;
+		$self->{details} = $self->{signatures}->[0]->{verify_details};
+		return;
 	}
 }
 
@@ -566,10 +547,8 @@ sub fetch_sender_policy
 	use Mail::DKIM::Policy;
 
 	# determine addresses found in the "From" and "Sender" headers
-	my $author = $self->message_originator;
-	$author &&= $author->address;
-	my $sender = $self->message_sender;
-	$sender &&= $sender->address;
+	my $author = $self->message_originator->address;
+	my $sender = $self->message_sender->address;
 
 	# fetch the policy
 	return Mail::DKIM::Policy->fetch(
@@ -591,21 +570,29 @@ terminators (same as the SMTP protocol).
 
   my $address = $dkim->message_originator;
 
-Returns the "originator address" found in the message. This is typically
-the (first) name and email address found in the From: header. The returned
-object is of type Mail::Address. To get just the email address part, do:
+Returns the "originator address" found in the message, as a
+L<Mail::Address> object.
+This is typically the (first) name and email address found in the
+From: header. If there is no From: header,
+then an empty L<Mail::Address> object is returned.
+
+To get just the email address part, do:
 
   my $email = $dkim->message_originator->address;
 
+See also L</"message_sender()">.
+
 =head2 message_sender() - access the "From" or "Sender" header
 
   my $address = $dkim->message_sender;
 
-Returns the "sender" found in the message. This is typically the (first)
-name and email address found in the Sender: header. If there is no Sender:
-header, it is the first name and email address in the From: header.
-The returned object is of type Mail::Address, so to get just the email
-address part, do:
+Returns the "sender" found in the message, as a L<Mail::Address> object.
+This is typically the (first) name and email address found in the
+Sender: header. If there is no Sender: header, it is the first name and
+email address in the From: header. If neither header is present,
+then an empty L<Mail::Address> object is returned.
+
+To get just the email address part, do:
 
   my $email = $dkim->message_sender->address;
 
@@ -671,7 +658,7 @@ The following are possible results from the result_detail() method:
   invalid (missing q tag)
   invalid (missing d tag)
   invalid (missing s tag)
-  invalid (unsupported v=0.1 tag)
+  invalid (unsupported version 0.1)
   invalid (no public key available)
   invalid (public key: unsupported version)
   invalid (public key: unsupported key type)
